@@ -7,137 +7,9 @@
 
 #include <fbzmq/zmq/Socket.h>
 
-#ifdef FOLLY_HAS_COROUTINES
-#include <folly/experimental/coro/Baton.h>
-#endif
 #include <folly/fibers/EventBaseLoopController.h>
-#include <folly/io/async/EventHandler.h>
 #include <folly/net/NetworkSocket.h>
 
-namespace {
-
-#ifdef FOLLY_HAS_COROUTINES
-//
-// ZMQ exposes special FD that in a sense similar to eventFd. The ZMQ
-// library will post events on the FD making it readable or writable,
-// to indicate that corresponding ZMQ socket is ready for reading or writing.
-//
-// The following EventHandler grabs that special FD for ZMQ socket, then
-// adds it to EventBase for tracking. The event handler fires when socket
-// state changes.
-//
-// We integrate the handler with co-routines by posting on coro::baton when
-// the corresponding event happens.
-//
-class CoroZmqSocketReadyHandler : public folly::EventHandler {
- public:
-  CoroZmqSocketReadyHandler(
-      folly::EventBase* evb,
-      int fd,
-      // pointer to raw ZMQ socket, so we can zmq_getsockopt()
-      void* zmqRawSocket,
-      folly::coro::Baton& baton,
-      bool isRead)
-      : EventHandler(evb, folly::NetworkSocket::fromFd(fd)),
-        isRead_{isRead},
-        zmqRawSocket_{zmqRawSocket},
-        baton_{baton} {
-    registerHandler(
-        (isRead_ ? EventHandler::READ : EventHandler::WRITE) |
-        EventHandler::PERSIST);
-  }
-
-  // we set this if ZMQ layer passes us an error
-  int errorCode{0};
-
-  void
-  handlerReady(uint16_t events) noexcept override {
-    if (!(events & (EventHandler::READ | EventHandler::WRITE))) {
-      return;
-    }
-
-    uint32_t zmqEvents{0};
-    size_t zmqEventsLen = sizeof(zmqEvents);
-
-    errorCode =
-        zmq_getsockopt(zmqRawSocket_, ZMQ_EVENTS, &zmqEvents, &zmqEventsLen);
-
-    // if ready or got error, notify the sleepr
-    if ((errorCode != 0) ||
-        (zmqEvents & (isRead_ ? ZMQ_POLLIN : ZMQ_POLLOUT))) {
-      unregisterHandler();
-      baton_.get().post();
-    }
-    // otherwise, we keep handler registered
-  }
-
- private:
-  bool isRead_{true};
-  void* zmqRawSocket_{nullptr};
-  std::reference_wrapper<folly::coro::Baton> baton_;
-};
-#endif
-
-//
-// ZMQ exposes special FD that in a sense similar to eventFd. The ZMQ
-// library will post events on the FD making it readable or writable,
-// to indicate that corresponding ZMQ socket is ready for reading or writing.
-//
-// The following EventHandler grabs that special FD for ZMQ socket, then
-// adds it to EventBase for tracking. The event handler fires when socket
-// state changes.
-//
-// We integrate the handler with folly::dibers by posting on fibers::baton when
-// the corresponding event happens.
-//
-class FiberZmqSocketReadyHandler : public folly::EventHandler {
- public:
-  FiberZmqSocketReadyHandler(
-      folly::EventBase* evb,
-      int fd,
-      // pointer to raw ZMQ socket, so we can zmq_getsockopt()
-      void* zmqRawSocket,
-      folly::fibers::Baton& baton,
-      bool isRead)
-      : EventHandler(evb, folly::NetworkSocket::fromFd(fd)),
-        isRead_{isRead},
-        zmqRawSocket_{zmqRawSocket},
-        baton_{baton} {
-    registerHandler(
-        (isRead_ ? EventHandler::READ : EventHandler::WRITE) |
-        EventHandler::PERSIST);
-  }
-
-  // we set this if ZMQ layer passes us an error
-  int errorCode{0};
-
-  void
-  handlerReady(uint16_t events) noexcept override {
-    if (!(events & (EventHandler::READ | EventHandler::WRITE))) {
-      return;
-    }
-
-    uint32_t zmqEvents{0};
-    size_t zmqEventsLen = sizeof(zmqEvents);
-    errorCode =
-        zmq_getsockopt(zmqRawSocket_, ZMQ_EVENTS, &zmqEvents, &zmqEventsLen);
-    // if ready or got error, notify the sleepr
-    if ((errorCode != 0) ||
-        (zmqEvents & (isRead_ ? ZMQ_POLLIN : ZMQ_POLLOUT))) {
-      unregisterHandler();
-      baton_.get().post();
-    }
-    // otherwise, we keep handler registered
-  }
-
- private:
-  bool isRead_{true};
-  void* zmqRawSocket_{nullptr};
-  std::reference_wrapper<folly::fibers::Baton> baton_;
-};
-
-} // namespace
-
 namespace fbzmq {
 namespace detail {
 
@@ -147,14 +19,16 @@ SocketImpl::SocketImpl(
     Context& ctx,
     folly::Optional<IdentityString> identity,
     folly::Optional<KeyPair> keyPair,
-    NonblockingFlag isNonblocking)
+    NonblockingFlag isNonblocking,
+    folly::EventBase* evb)
     : SocketImpl(
           type,
           isServer,
           ctx.ptr_,
           std::move(identity),
           std::move(keyPair),
-          isNonblocking) {}
+          isNonblocking,
+          evb) {}
 
 SocketImpl::SocketImpl(
     int type,
@@ -162,10 +36,13 @@ SocketImpl::SocketImpl(
     void* ctxPtr,
     folly::Optional<IdentityString> identity,
     folly::Optional<KeyPair> keyPair,
-    NonblockingFlag isNonblocking)
-    : ptr_(zmq_socket(ctxPtr, type)),
+    NonblockingFlag isNonblocking,
+    folly::EventBase* evb)
+    : folly::EventHandler(),
+      ptr_(zmq_socket(ctxPtr, type)),
       ctxPtr_{ctxPtr},
-      keyPair_(std::move(keyPair)) {
+      keyPair_(std::move(keyPair)),
+      evb_(evb) {
   CHECK(ctxPtr);
   CHECK(ptr_) << Error();
 
@@ -173,6 +50,8 @@ SocketImpl::SocketImpl(
     baseFlags_ |= ZMQ_DONTWAIT;
   }
 
+  initHandlerHelper();
+
   // Enable ZMQ_IPV6 by default on the sockets
   const int ipv6Enable = 1;
   setSockOpt(ZMQ_IPV6, &ipv6Enable, sizeof(int)).value();
@@ -211,11 +90,14 @@ SocketImpl::~SocketImpl() noexcept {
 }
 
 SocketImpl::SocketImpl(SocketImpl&& other) noexcept
-    : baseFlags_(other.baseFlags_),
+    : folly::EventHandler(),
+      baseFlags_(other.baseFlags_),
       ptr_(other.ptr_),
       ctxPtr_(other.ctxPtr_),
-      keyPair_(std::move(other.keyPair_)) {
+      keyPair_(std::move(other.keyPair_)),
+      evb_(other.evb_) {
   other.ptr_ = nullptr;
+  initHandlerHelper();
 }
 
 SocketImpl&
@@ -224,12 +106,28 @@ SocketImpl::operator=(SocketImpl&& other) noexcept {
   ptr_ = other.ptr_;
   ctxPtr_ = other.ctxPtr_;
   keyPair_ = std::move(other.keyPair_);
-
+  evb_ = other.evb_;
   other.ptr_ = nullptr;
-
+  initHandlerHelper();
   return *this;
 }
 
+void
+SocketImpl::initHandlerHelper() noexcept {
+  if (not evb_) {
+    return;
+  }
+
+  CHECK(baseFlags_ & ZMQ_DONTWAIT)
+      << "Socket must be set in non-blocking mode for async read/writes";
+  // Initialize event handler
+  int socketFd{-1};
+  size_t fdLen = sizeof(socketFd);
+  const auto rc = zmq_getsockopt(ptr_, ZMQ_FD, &socketFd, &fdLen);
+  CHECK_EQ(0, rc) << "Can't get fd for socket. " << Error();
+  initHandler(evb_, folly::NetworkSocket::fromFd(socketFd));
+}
+
 folly::Expected<folly::Unit, Error>
 SocketImpl::setSockOpt(int option, const void* optval, size_t len) noexcept {
   const int rc = zmq_setsockopt(ptr_, option, optval, len);
@@ -296,106 +194,162 @@ SocketImpl::close() noexcept {
   if (not ptr_) {
     return;
   }
-  if (fiberBaton_) {
-    // Notice this does not wake up blocking read (recvOne...)
-    // when socket close happen
-    fiberBaton_->post();
-  }
+  // Unregister handler from event base
+  unregisterHandler();
+
+  // Unblock awaited reads or writes if any
+#if FOLLY_HAS_COROUTINES
+  coroReadBaton_.post();
+  coroWriteBaton_.post();
+#endif
+  fiberReadBaton_.post();
+  fiberWriteBaton_.post();
+
   const int rc = zmq_close(ptr_);
   CHECK_EQ(0, rc) << zmq_strerror(zmq_errno());
   ptr_ = nullptr;
 }
 
-#ifdef FOLLY_HAS_COROUTINES
-folly::coro::Task<folly::Expected<folly::Unit, Error>>
-SocketImpl::coroWaitToRecv(folly::EventBase* evb) {
-  co_return co_await coroWaitImpl(evb, WaitReason::RECV);
-}
+#if FOLLY_HAS_COROUTINES
 
-folly::coro::Task<folly::Expected<folly::Unit, Error>>
-SocketImpl::coroWaitToSend(folly::EventBase* evb) {
-  co_return co_await coroWaitImpl(evb, WaitReason::SEND);
-}
+folly::coro::Task<folly::Expected<Message, Error>>
+SocketImpl::recvOneCoro() {
+  CHECK(evb_);
 
-folly::coro::Task<folly::Expected<folly::Unit, Error>>
-SocketImpl::coroWaitImpl(folly::EventBase* evb, WaitReason reason) {
-  folly::coro::Baton baton;
-  int zmqFd{-1};
-  size_t fdLen = sizeof(zmqFd);
-
-  const auto rc = zmq_getsockopt(ptr_, ZMQ_FD, &zmqFd, &fdLen);
-
-  if (rc != 0) {
-    co_return folly::makeUnexpected(Error());
+  // Try immediate recv
+  auto ret = recv(baseFlags_);
+  if (not ret.hasError() or ret.error().errNum != EAGAIN) {
+    co_return ret;
   }
 
-  CoroZmqSocketReadyHandler zh(
-      evb, zmqFd, ptr_, baton, reason == WaitReason::RECV);
-  // when zmqFd becomes readable, the event handler will post
-  // on the baton. Until then, this coroutine will be suspended
-  co_await baton;
-  if (zh.errorCode) {
-    co_return folly::makeUnexpected(Error(zh.errorCode));
+  // Error was EAGAIN - Wait for socket to become readable, then perform read
+  co_await coroWaitImpl(true /* isReadElseWrite */);
+  co_return recv(baseFlags_);
+}
+
+folly::coro::Task<folly::Expected<size_t, Error>>
+SocketImpl::sendOneCoro(Message msg) {
+  CHECK(evb_);
+
+  // Wait for socket to become writable, then perform write
+  // NOTE: This is different from read path as we will always
+  // get write event on epoll if we haven't written before.
+  co_await coroWaitImpl(false /* isReadElseWrite */);
+  co_return send(std::move(msg), baseFlags_);
+}
+
+folly::coro::Task<void>
+SocketImpl::coroWaitImpl(bool isReadElseWrite) noexcept {
+  folly::coro::Baton* waitBaton{nullptr};
+  if (isReadElseWrite) {
+    waitBaton = &coroReadBaton_;
+    waitEvents_ |= EventHandler::READ;
+  } else {
+    waitBaton = &coroWriteBaton_;
+    waitEvents_ |= EventHandler::WRITE;
   }
-  co_return folly::unit;
+  CHECK(waitBaton);
+  waitBaton->reset(); // Reset baton
+  registerHandler(waitEvents_ | EventHandler::PERSIST);
+  co_await *waitBaton;
+  co_return;
 }
 #endif
 
-folly::Expected<folly::Unit, Error>
-SocketImpl::fiberWaitToRecv() {
-  return fiberWaitImpl(WaitReason::RECV);
-}
-
-folly::Expected<folly::Unit, Error>
-SocketImpl::fiberWaitToSend() {
-  return fiberWaitImpl(WaitReason::SEND);
-}
-
-folly::Expected<folly::Unit, Error>
-SocketImpl::fiberWaitImpl(WaitReason reason) {
+bool
+SocketImpl::fiberWaitImpl(
+    bool isReadElseWrite,
+    folly::Optional<std::chrono::milliseconds> timeout) noexcept {
   CHECK(folly::fibers::onFiber()) << "Not on fiber!";
 
-  auto& fm = folly::fibers::FiberManager::getFiberManager();
-  auto& lc =
-      static_cast<folly::fibers::EventBaseLoopController&>(fm.loopController());
-  auto evb = &lc.getEventBase()->getEventBase();
+  folly::fibers::Baton* waitBaton{&fiberReadBaton_};
+  uint16_t waitEvent{EventHandler::READ};
+  if (not isReadElseWrite) {
+    waitBaton = &fiberWriteBaton_;
+    waitEvent = EventHandler::WRITE;
+  }
+  CHECK(waitBaton);
+  waitEvents_ |= waitEvent;
+  waitBaton->reset(); // Reset baton
+  registerHandler(waitEvents_ | EventHandler::PERSIST);
+  if (timeout.has_value()) {
+    const auto hasEvent = waitBaton->timed_wait(timeout.value());
+    if (not hasEvent) {
+      // Unregister wait-event
+      waitEvents_ &= ~waitEvent;
+      if (not waitEvents_) {
+        unregisterHandler();
+      } else {
+        registerHandler(waitEvents_);
+      }
+      return false;
+    }
+  } else {
+    waitBaton->wait();
+  }
+  return true;
+}
 
-  folly::fibers::Baton baton;
-  fiberBaton_ = &baton;
-  int zmqFd{-1};
-  size_t fdLen = sizeof(zmqFd);
+void
+SocketImpl::handlerReady(uint16_t events) noexcept {
+  // Event must be of our interest
+  CHECK(events & EventHandler::READ_WRITE)
+      << "Received unknown event(s): " << events;
 
-  const auto rc = zmq_getsockopt(ptr_, ZMQ_FD, &zmqFd, &fdLen);
+  // Retrieve zmq events
+  uint32_t zmqEvents{0};
+  size_t zmqEventsLen = sizeof(zmqEvents);
+  auto err = zmq_getsockopt(ptr_, ZMQ_EVENTS, &zmqEvents, &zmqEventsLen);
+  CHECK_EQ(0, err) << "Got error while reading events from zmq socket";
 
-  if (rc != 0) {
-    return folly::makeUnexpected(Error());
+  const auto prevWaitEvents = waitEvents_;
+
+  // If read events, wake up reader
+  if ((waitEvents_ & EventHandler::READ) and (zmqEvents & ZMQ_POLLIN)) {
+    waitEvents_ &= ~EventHandler::READ; // Remove read event
+    fiberReadBaton_.post();
+#if FOLLY_HAS_COROUTINES
+    coroReadBaton_.post();
+#endif
   }
 
-  FiberZmqSocketReadyHandler zh(
-      evb, zmqFd, ptr_, baton, reason == WaitReason::RECV);
+  // If write events, wake up writer
+  if ((waitEvents_ & EventHandler::WRITE) and (zmqEvents & ZMQ_POLLOUT)) {
+    waitEvents_ &= ~EventHandler::WRITE; // Remove write event
+    fiberWriteBaton_.post();
+#if FOLLY_HAS_COROUTINES
+    coroWriteBaton_.post();
+#endif
+  }
 
-  // when zmqFd becomes readable, the event handler will post
-  // on the baton. Until then, this coroutine will be suspended
-  baton.wait();
-  fiberBaton_ = nullptr;
-  if (zh.errorCode) {
-    return folly::makeUnexpected(Error(zh.errorCode));
+  // Un-register or update handler for waiting on remaining event
+  if (not waitEvents_) {
+    unregisterHandler();
+  } else if (waitEvents_ != prevWaitEvents) {
+    registerHandler(waitEvents_); // Update handler with new wait events
   }
-  return folly::unit;
 }
 
 folly::Expected<size_t, Error>
-SocketImpl::sendOne(Message msg) const noexcept {
+SocketImpl::sendOne(Message msg) noexcept {
+  // Wait only if sending first part and evb is present
+  if (not isSendingMore_ && evb_) {
+    fiberWaitImpl(false /* isReadElseWrite */, folly::none);
+  }
   return send(std::move(msg), baseFlags_);
 }
 
 folly::Expected<size_t, Error>
-SocketImpl::sendMore(Message msg) const noexcept {
+SocketImpl::sendMore(Message msg) noexcept {
+  // Wait only if sending first part and evb is present
+  if (not isSendingMore_ && evb_) {
+    fiberWaitImpl(false /* isReadElseWrite */, folly::none);
+  }
   return send(std::move(msg), baseFlags_ | ZMQ_SNDMORE);
 }
 
 folly::Expected<size_t, Error>
-SocketImpl::sendMultiple(std::vector<Message> const& msgs, bool hasMore) const {
+SocketImpl::sendMultiple(std::vector<Message> const& msgs, bool hasMore) {
   size_t size{0};
   if (msgs.empty()) {
     return 0;
@@ -416,11 +370,29 @@ SocketImpl::sendMultiple(std::vector<Message> const& msgs, bool hasMore) const {
   return size += last.value();
 }
 
+folly::Expected<Message, Error>
+SocketImpl::recvAsync(
+    folly::Optional<std::chrono::milliseconds> timeout) noexcept {
+  auto ret = recv(baseFlags_);
+  if (not ret.hasError() or ret.error().errNum != EAGAIN) {
+    return ret;
+  }
+
+  // Error was EAGAIN - Wait for socket to become readable, then perform read
+  if (not fiberWaitImpl(true /* isReadElseWrite */, timeout)) {
+    return folly::makeUnexpected(Error(EAGAIN, "fiber recv timeout"));
+  }
+  return recv(baseFlags_);
+}
+
 folly::Expected<Message, Error>
 SocketImpl::recvOne(folly::Optional<std::chrono::milliseconds>
-                        timeout /* = folly::none */) const noexcept {
+                        timeout /* = folly::none */) noexcept {
   // ignore timeout if non-blocking
   if (baseFlags_ & ZMQ_DONTWAIT) {
+    if (evb_) {
+      return recvAsync(timeout);
+    }
     return recv(baseFlags_);
   }
 
@@ -543,7 +515,8 @@ SocketImpl::disconnect(SocketUrl addr) noexcept {
 }
 
 folly::Expected<size_t, Error>
-SocketImpl::send(Message msg, int flags) const noexcept {
+SocketImpl::send(Message msg, int flags) noexcept {
+  isSendingMore_ = flags & ZMQ_SNDMORE;
   while (true) {
     const int n = zmq_msg_send(&(msg.msg_), ptr_, flags);
     if (n >= 0) {
@@ -572,7 +545,7 @@ SocketImpl::delServerKey(SocketUrl server) noexcept {
 }
 
 folly::Expected<Message, Error>
-SocketImpl::recv(int flags) const noexcept {
+SocketImpl::recv(int flags) noexcept {
   Message msg;
   while (true) {
     const int n = zmq_msg_recv(&(msg.msg_), ptr_, flags);
