@@ -17,8 +17,13 @@
 #include <folly/fibers/Baton.h>
 #include <folly/fibers/EventBaseLoopController.h>
 #include <folly/fibers/FiberManager.h>
+#include <folly/init/Init.h>
 #include <folly/io/async/AsyncTimeout.h>
 
+#if FOLLY_HAS_COROUTINES
+#include <folly/experimental/coro/Sleep.h>
+#endif
+
 using namespace std::chrono_literals;
 
 using apache::thrift::CompactSerializer;
@@ -182,12 +187,14 @@ TEST(Socket, SingleMessage) {
   EXPECT_EQ(str, rcvd.value().read<std::string>().value());
 }
 
-#ifdef FOLLY_HAS_COROUTINES
+#if FOLLY_HAS_COROUTINES
 TEST(Socket, CoroSingleMessage) {
-  fbzmq::Context ctx;
-  fbzmq::Socket<ZMQ_REQ, fbzmq::ZMQ_CLIENT> req(ctx);
-  fbzmq::Socket<ZMQ_REP, fbzmq::ZMQ_SERVER> rep(ctx);
   folly::EventBase evb;
+  fbzmq::Context ctx;
+  fbzmq::Socket<ZMQ_REQ, fbzmq::ZMQ_CLIENT> req(
+      ctx, folly::none, folly::none, NonblockingFlag{true}, &evb);
+  fbzmq::Socket<ZMQ_REP, fbzmq::ZMQ_SERVER> rep(
+      ctx, folly::none, folly::none, NonblockingFlag{true}, &evb);
 
   rep.bind(fbzmq::SocketUrl{"inproc://test"}).value();
   req.connect(fbzmq::SocketUrl{"inproc://test"}).value();
@@ -196,14 +203,13 @@ TEST(Socket, CoroSingleMessage) {
 
   auto writer = [&evb, &req, &str]() -> folly::coro::Task<folly::Unit> {
     auto msg = fbzmq::Message::from(str).value();
-    co_await req.coroWaitToSend(&evb);
-    req.sendOne(msg);
+    auto ret = co_await req.sendOneCoro(msg);
+    EXPECT_TRUE(ret.hasValue());
     co_return folly::unit;
   };
 
   auto reader = [&evb, &rep, &str]() -> folly::coro::Task<bool> {
-    co_await rep.coroWaitToRecv(&evb);
-    auto rcvd = rep.recvOne();
+    auto rcvd = co_await rep.recvOneCoro();
     co_return str == rcvd.value().read<std::string>().value();
   };
 
@@ -214,15 +220,71 @@ TEST(Socket, CoroSingleMessage) {
   EXPECT_TRUE(futReader.isReady());
   EXPECT_TRUE(futReader.value());
 }
+
+TEST(Socket, CoroPubSub) {
+  folly::EventBase evb;
+  fbzmq::Context ctx;
+  fbzmq::Socket<ZMQ_SUB, fbzmq::ZMQ_CLIENT> sub(
+      ctx, folly::none, folly::none, NonblockingFlag{true}, &evb);
+  fbzmq::Socket<ZMQ_PUB, fbzmq::ZMQ_SERVER> pub(
+      ctx, folly::none, folly::none, NonblockingFlag{true}, &evb);
+
+  pub.bind(fbzmq::SocketUrl{"inproc://test"}).value();
+  sub.connect(fbzmq::SocketUrl{"inproc://test"}).value();
+  sub.setSockOpt(ZMQ_SUBSCRIBE, "", 0).value();
+
+  const auto str = genRandomStr(1024);
+
+  auto writer = [&pub, &str]() -> folly::coro::Task<void> {
+    auto msg = fbzmq::Message::from(str).value();
+    size_t count{128};
+    while (count--) {
+      VLOG(1) << "Writer sending message " << count;
+      auto ret = co_await pub.sendOneCoro(msg);
+      EXPECT_TRUE(ret.hasValue());
+      co_await folly::coro::sleep(
+          std::chrono::milliseconds(folly::Random::rand32(0, 10)));
+    }
+    auto ret =
+        co_await pub.sendOneCoro(fbzmq::Message::from(std::string("")).value());
+    EXPECT_TRUE(ret.hasValue());
+    co_return;
+  };
+
+  auto reader = [&sub, &str]() -> folly::coro::Task<size_t> {
+    size_t count{0};
+    while (true) {
+      auto rcvd = co_await sub.recvOneCoro();
+      VLOG(1) << "Reader received message " << count;
+      auto data = rcvd.value().read<std::string>().value();
+      if (data.size() == 0) {
+        break;
+      }
+      ++count;
+      EXPECT_EQ(str, data);
+    }
+    co_return count;
+  };
+
+  auto futReader = reader().scheduleOn(&evb).start();
+  auto futWriter = writer().scheduleOn(&evb).start();
+  evb.loop();
+
+  EXPECT_TRUE(futReader.isReady());
+  EXPECT_EQ(128, futReader.value());
+  EXPECT_TRUE(futWriter.isReady());
+}
 #endif
 
 TEST(Socket, FiberSingleMessage) {
   using namespace folly::fibers;
 
-  fbzmq::Context ctx;
-  fbzmq::Socket<ZMQ_REQ, fbzmq::ZMQ_CLIENT> req(ctx);
-  fbzmq::Socket<ZMQ_REP, fbzmq::ZMQ_SERVER> rep(ctx);
   folly::EventBase evb;
+  fbzmq::Context ctx;
+  fbzmq::Socket<ZMQ_REQ, fbzmq::ZMQ_CLIENT> req(
+      ctx, folly::none, folly::none, NonblockingFlag{true}, &evb);
+  fbzmq::Socket<ZMQ_REP, fbzmq::ZMQ_SERVER> rep(
+      ctx, folly::none, folly::none, NonblockingFlag{true}, &evb);
 
   rep.bind(fbzmq::SocketUrl{"inproc://test"}).value();
   req.connect(fbzmq::SocketUrl{"inproc://test"}).value();
@@ -237,13 +299,11 @@ TEST(Socket, FiberSingleMessage) {
   auto sender = fm->addTaskFuture([&req, &str]() {
     VLOG(1) << "Start sender fiber";
     auto msg = fbzmq::Message::from(str).value();
-    req.fiberWaitToSend();
     req.sendOne(msg);
   });
 
   auto receiver = fm->addTaskFuture([&rep, &str]() {
     VLOG(1) << "Start receiver fiber";
-    rep.fiberWaitToRecv();
     auto rcvd = rep.recvOne();
     EXPECT_EQ(str, rcvd.value().read<std::string>().value());
   });
@@ -258,15 +318,15 @@ TEST(Socket, FiberSingleMessage) {
 TEST(Socket, FiberPubSubMessage) {
   using namespace folly::fibers;
 
+  folly::EventBase evb;
   fbzmq::Context ctx;
   fbzmq::Socket<ZMQ_SUB, fbzmq::ZMQ_CLIENT> sub(
-      ctx, folly::none, folly::none, fbzmq::NonblockingFlag{true});
+      ctx, folly::none, folly::none, fbzmq::NonblockingFlag{true}, &evb);
   fbzmq::Socket<ZMQ_PUB, fbzmq::ZMQ_SERVER> pub(
-      ctx, folly::none, folly::none, fbzmq::NonblockingFlag{true});
-  folly::EventBase evb;
+      ctx, folly::none, folly::none, fbzmq::NonblockingFlag{true}, &evb);
 
-  pub.bind(fbzmq::SocketUrl{"ipc:///tmp/test"}).value();
-  sub.connect(fbzmq::SocketUrl{"ipc:///tmp/test"}).value();
+  pub.bind(fbzmq::SocketUrl{"inproc:///tmp/test"}).value();
+  sub.connect(fbzmq::SocketUrl{"inproc:///tmp/test"}).value();
   sub.setSockOpt(ZMQ_SUBSCRIBE, "", 0).value();
 
   auto fm = std::make_unique<FiberManager>(
@@ -277,37 +337,42 @@ TEST(Socket, FiberPubSubMessage) {
   const auto str = genRandomStr(1024);
 
   // receiver needs to drain all messages before put back to wait
-  auto receiver = fm->addTaskFuture([&sub]() {
+  auto receiver = fm->addTaskFuture([&sub]() noexcept {
+    LOG(INFO) << "Receiver starting";
     auto rcvdMsgCnt = 0;
     while (true) {
-      sub.fiberWaitToRecv();
-      auto rcvd = sub.drain().value();
+      auto rcvd = sub.recvOne();
       // if sender sends too fast we will receive >1 messages here
-      rcvdMsgCnt += rcvd.size();
-      VLOG(1) << "Receiver got " << rcvdMsgCnt << " msgs";
+      ++rcvdMsgCnt;
+      VLOG(1) << "Receiver got msg " << rcvdMsgCnt;
       if (rcvdMsgCnt == 1024) {
         break;
       }
     }
+    LOG(INFO) << "Receiver terminating";
     SUCCEED();
   });
 
-  auto sender = fm->addTaskFuture([&pub, &str, &evb]() {
+  auto sender = fm->addTaskFuture([&pub, &str, &evb]() noexcept {
+    LOG(INFO) << "Sender starting";
     for (int i = 0; i < 1024; i++) {
       auto msg = fbzmq::Message::from(str).value();
-      pub.fiberWaitToSend();
+      VLOG(1) << "Sender sending msg " << (i + 1);
       pub.sendOne(msg);
       // random wait before send next msg
       folly::fibers::Baton bt;
       auto timeout = folly::AsyncTimeout::schedule(
-          std::chrono::milliseconds(folly::Random::rand32(0, 20)),
+          std::chrono::milliseconds(folly::Random::rand32(0, 10)),
           evb,
           [&bt]() noexcept { bt.post(); });
       bt.wait();
+      LOG(INFO) << "Sender terminating";
     }
   });
 
+  LOG(INFO) << "Starting eventbase loop";
   evb.loop();
+  LOG(INFO) << "eventbase loop terminated";
 
   // wait for fiber to finish
   std::move(sender).get();
@@ -317,10 +382,10 @@ TEST(Socket, FiberPubSubMessage) {
 TEST(Socket, FiberSubSocketClose) {
   using namespace folly::fibers;
 
+  folly::EventBase evb;
   fbzmq::Context ctx;
   fbzmq::Socket<ZMQ_SUB, fbzmq::ZMQ_CLIENT> sub(
-      ctx, folly::none, folly::none, fbzmq::NonblockingFlag{true});
-  folly::EventBase evb;
+      ctx, folly::none, folly::none, fbzmq::NonblockingFlag{true}, &evb);
 
   sub.connect(fbzmq::SocketUrl{"inproc://test"}).value();
 
@@ -330,10 +395,8 @@ TEST(Socket, FiberSubSocketClose) {
       .attachEventBase(evb);
 
   // both will block on waitToRead
-  auto client = fm->addTaskFuture([&sub]() {
-    sub.fiberWaitToRecv();
-    EXPECT_TRUE(sub.recvOne().hasError());
-  });
+  auto client =
+      fm->addTaskFuture([&sub]() { EXPECT_TRUE(sub.recvOne().hasError()); });
 
   fm->addTask([&sub, &evb]() {
     // wait for 50ms before close the socket
@@ -341,7 +404,7 @@ TEST(Socket, FiberSubSocketClose) {
     auto timeout = folly::AsyncTimeout::schedule(
         std::chrono::milliseconds(50), evb, [&bt]() noexcept { bt.post(); });
     bt.wait();
-    // close both sockets
+    // close socket. Reader will get unblocked
     sub.close();
   });
 
@@ -351,6 +414,115 @@ TEST(Socket, FiberSubSocketClose) {
   std::move(client).get();
 }
 
+TEST(Socket, FiberSendRecvMultiple) {
+  using namespace folly::fibers;
+
+  folly::EventBase evb;
+  fbzmq::Context ctx;
+  fbzmq::Socket<ZMQ_PUB, fbzmq::ZMQ_SERVER> pub(
+      ctx, folly::none, folly::none, NonblockingFlag{true}, &evb);
+  fbzmq::Socket<ZMQ_SUB, fbzmq::ZMQ_CLIENT> sub(
+      ctx, folly::none, folly::none, NonblockingFlag{true}, &evb);
+
+  pub.bind(fbzmq::SocketUrl{"inproc://test"}).value();
+  sub.connect(fbzmq::SocketUrl{"inproc://test"}).value();
+  sub.setSockOpt(ZMQ_SUBSCRIBE, "", 0).value();
+
+  auto fm = std::make_unique<FiberManager>(
+      std::make_unique<EventBaseLoopController>());
+  static_cast<EventBaseLoopController&>(fm->loopController())
+      .attachEventBase(evb);
+
+  auto sender = fm->addTaskFuture([&evb, &pub]() {
+    VLOG(1) << "Start sender fiber";
+    std::vector<fbzmq::Message> msgs{
+        fbzmq::Message::from(genRandomStr(1024)).value(),
+        fbzmq::Message::from(genRandomStr(1024)).value(),
+        fbzmq::Message::from(genRandomStr(1024)).value(),
+        fbzmq::Message::from(genRandomStr(1024)).value()};
+
+    // Templatized send
+    {
+      LOG(INFO) << "Sending multiple messages as variadic args";
+      auto ret = pub.sendMultiple(msgs.at(0), msgs.at(1), msgs.at(2));
+      EXPECT_TRUE(ret.hasValue());
+    }
+
+    // Wait
+    LOG(INFO) << "Sender sleeping for 100ms";
+    folly::fibers::Baton bt;
+    auto timeout = folly::AsyncTimeout::schedule(
+        std::chrono::milliseconds(100), evb, [&bt]() noexcept { bt.post(); });
+    bt.wait();
+
+    // Send multiple with vector
+    {
+      LOG(INFO) << "Sending multiple messages as vector - continued";
+      auto ret = pub.sendMultiple(msgs, true /* hasMore */);
+      EXPECT_TRUE(ret.hasValue());
+    }
+    {
+      LOG(INFO) << "Sending multiple messages as vector - done";
+      auto ret = pub.sendMultiple(msgs, false /* hasMore */);
+      EXPECT_TRUE(ret.hasValue());
+    }
+    LOG(INFO) << "Sender terminating";
+  });
+
+  auto receiver = fm->addTaskFuture([&sub]() {
+    VLOG(1) << "Start receiver fiber";
+
+    fbzmq::Message msg1, msg2, msg3;
+    auto ret = sub.recvMultiple(msg1, msg2, msg3);
+    LOG(INFO) << "Received messages with variadic args";
+    EXPECT_TRUE(ret.hasValue());
+    EXPECT_EQ(1024, msg1.size());
+    EXPECT_EQ(1024, msg2.size());
+    EXPECT_EQ(1024, msg3.size());
+
+    // NOTE: We will receive all 8 parts in a single shot
+    auto msgs = sub.recvMultiple();
+    LOG(INFO) << "Received messages with vector list";
+    ASSERT_TRUE(msgs.hasValue());
+    EXPECT_EQ(8, msgs.value().size());
+
+    LOG(INFO) << "Receiver terminating";
+  });
+
+  evb.loop();
+
+  // wait for fiber to finish
+  std::move(sender).get();
+  std::move(receiver).get();
+}
+
+TEST(Socket, FiberRecvTimeout) {
+  using namespace folly::fibers;
+
+  folly::EventBase evb;
+  auto fm = std::make_unique<FiberManager>(
+      std::make_unique<EventBaseLoopController>());
+  static_cast<EventBaseLoopController&>(fm->loopController())
+      .attachEventBase(evb);
+
+  fbzmq::Context ctx;
+  fbzmq::Socket<ZMQ_SUB, fbzmq::ZMQ_CLIENT> sub(
+      ctx, folly::none, folly::none, fbzmq::NonblockingFlag{true}, &evb);
+  sub.connect(fbzmq::SocketUrl{"inproc://test"}).value();
+
+  // both will block on waitToRead
+  auto client = fm->addTaskFuture([&sub]() {
+    auto recvd = sub.recvOne(std::chrono::milliseconds(100));
+    ASSERT_TRUE(recvd.hasError());
+    EXPECT_EQ(EAGAIN, recvd.error().errNum);
+  });
+
+  evb.loop();
+
+  // wait for fiber to finish
+  std::move(client).get();
+}
+
 //
 // Receive within timeout: blocking
 //
@@ -494,25 +666,25 @@ TEST(Socket, ThriftSerDeser) {
 
   // The thrift "value" object
   fbzmq::test::TestValue clientVal;
-  clientVal.value = "test1234";
+  *clientVal.value() = "test1234";
 
   auto rawClientVal = fbzmq::util::writeThriftObj(clientVal, serializer);
 
   // The wrapper we use to store something in KV database
   fbzmq::test::WrapperValue wrapperVal;
-  wrapperVal.version = 1000;
-  wrapperVal.value = *rawClientVal;
+  *wrapperVal.version() = 1000;
+  *wrapperVal.value() = *rawClientVal;
 
   auto rawWrapperVal = fbzmq::util::writeThriftObj(wrapperVal, serializer);
   auto newWrapperVal = fbzmq::util::readThriftObj<fbzmq::test::WrapperValue>(
       *(rawWrapperVal.get()), serializer);
 
-  EXPECT_EQ(newWrapperVal.version, 1000);
+  EXPECT_EQ(*newWrapperVal.version(), 1000);
 
   auto newClientVal = fbzmq::util::readThriftObj<fbzmq::test::TestValue>(
-      newWrapperVal.value, serializer);
+      *newWrapperVal.value(), serializer);
 
-  EXPECT_EQ(newClientVal.value, "test1234");
+  EXPECT_EQ(*newClientVal.value(), "test1234");
 }
 
 //
@@ -521,7 +693,7 @@ TEST(Socket, ThriftSerDeser) {
 TEST(Socket, ThriftSerDeserStr) {
   CompactSerializer serializer;
   fbzmq::test::TestValue clientVal;
-  clientVal.value = "hello world";
+  *clientVal.value() = "hello world";
 
   auto str = fbzmq::util::writeThriftObjStr(clientVal, serializer);
   auto obj =
@@ -551,6 +723,7 @@ TEST(Socket, SendRecvString) {
       {reinterpret_cast<void*>(*rep), 0, ZMQ_POLLIN, 0}};
   fbzmq::poll(pollItems).then([&serializer, &rep, &value](int) {
     rep.recvOne().then([&serializer, &value](fbzmq::Message&& rcvd) {
+      (void)&serializer; // silence clang -Wunused-lambda-capture
       auto rcvdValue = rcvd.read<std::string>();
       EXPECT_EQ(value, rcvdValue.value());
     });
@@ -572,7 +745,7 @@ TEST(Socket, SendRecvThriftObj) {
   // The thrift "value" object
   const auto str = genRandomStr(32768);
   fbzmq::test::TestValue testValue;
-  testValue.value = str;
+  *testValue.value() = str;
 
   fbzmq::Message::fromThriftObj(testValue, serializer)
       .then([&req](fbzmq::Message&& msg) { req.sendOne(std::move(msg)); });
@@ -582,7 +755,7 @@ TEST(Socket, SendRecvThriftObj) {
   fbzmq::poll(pollItems).then([&serializer, &rep, &str](int) {
     rep.recvOne().then([&serializer, &str](fbzmq::Message&& rcvd) {
       auto rcvdValue = rcvd.readThriftObj<fbzmq::test::TestValue>(serializer);
-      EXPECT_EQ(str, rcvdValue.value().value);
+      EXPECT_EQ(str, *rcvdValue.value().value());
     });
   });
 }
@@ -603,7 +776,7 @@ TEST(Socket, ThriftObjPingPong) {
     {
       const auto str = genRandomStr(32768);
       fbzmq::test::TestValue testValue;
-      testValue.value = str;
+      *testValue.value() = str;
 
       fbzmq::Message::fromThriftObj(testValue, serializer)
           .then([&req](fbzmq::Message&& msg) { req.sendOne(std::move(msg)); });
@@ -614,7 +787,7 @@ TEST(Socket, ThriftObjPingPong) {
         rep.recvOne().then([&serializer, &str](fbzmq::Message&& rcvd) {
           auto rcvdValue =
               rcvd.readThriftObj<fbzmq::test::TestValue>(serializer);
-          EXPECT_EQ(str, rcvdValue.value().value);
+          EXPECT_EQ(str, *rcvdValue.value().value());
         });
       });
     }
@@ -622,7 +795,7 @@ TEST(Socket, ThriftObjPingPong) {
     {
       const auto str = genRandomStr(32768);
       fbzmq::test::TestValue testValue;
-      testValue.value = str;
+      *testValue.value() = str;
 
       fbzmq::Message::fromThriftObj(testValue, serializer)
           .then([&rep](fbzmq::Message&& msg) { rep.sendOne(std::move(msg)); });
@@ -633,7 +806,7 @@ TEST(Socket, ThriftObjPingPong) {
         req.recvOne().then([&serializer, &str](fbzmq::Message&& rcvd) {
           auto rcvdValue =
               rcvd.readThriftObj<fbzmq::test::TestValue>(serializer);
-          EXPECT_EQ(str, rcvdValue.value().value);
+          EXPECT_EQ(str, *rcvdValue.value().value());
         });
       });
     }
@@ -660,7 +833,7 @@ TEST(Socket, ThriftObjectSendManyTwoThreads) {
                      .value()
                      .readThriftObj<fbzmq::test::TestValue>(serializer)
                      .value();
-      EXPECT_EQ(32768, obj.value.size());
+      EXPECT_EQ(32768, obj.value()->size());
     }
     server.unbind(fbzmq::SocketUrl{"inproc://test"}).value();
   });
@@ -673,7 +846,7 @@ TEST(Socket, ThriftObjectSendManyTwoThreads) {
     for (int i = 0; i < 1024; i++) {
       const auto str = genRandomStr(32768);
       fbzmq::test::TestValue testValue;
-      testValue.value = str;
+      *testValue.value() = str;
 
       fbzmq::Message::fromThriftObj(testValue, serializer)
           .then([&client](fbzmq::Message&& msg) {
@@ -1192,8 +1365,7 @@ int
 main(int argc, char* argv[]) {
   // Parse command line flags
   testing::InitGoogleTest(&argc, argv);
-  gflags::ParseCommandLineFlags(&argc, &argv, true);
-  google::InitGoogleLogging(argv[0]);
+  folly::init(&argc, &argv);
 
   // init sodium security library
   if (::sodium_init() == -1) {
